package Hashlocker;

use strict;
use Baseobject;
use Lock;
use Fcntl;


my $version ='$Id: Hashlocker.pm 4553 2004-08-30 10:43:35Z anton $ ';

our $VERSION ="1.5";

our @ISA =('Baseobject' );

sub new
{
  my $pkg=shift;
  my $self= bless (new Baseobject(@_),$pkg);

  my $defaultpath='.';
  $defaultpath=$main::localstatedir if (defined($main::localstatedir));

  $self->allows(
		{
		  'db'=>$self->gendbfilename(),
		  'dbpath'=>$defaultpath,
		  'umask'   =>umask,
		  'fileperm'=>0600,
		  'filer'=>'NDBM',
		  'nolock'=>0,
		  'timeout'=>undef,
		  'keeplocked'=>0,
		  'lock'=>0,
		}
		);

  $self->field('name','Hashlocker');
  $self->fields(@_);

  my $extension = $self->field('db');
  my $hasextension =$extension =~ s/(^.*)\.//;
  my $db=$1;

  if ($hasextension)
  {
    $self->field('db',$db);

    if ($extension =~  /^db$|^pag$|^dir$/)
    {
      $self->field('filer','NDBM')
	}
    else
    {
      $self->field('filer',$extension);
    }
  }


  if ($self->field('dbpath'))
  {
    unless ( -d $self->field('dbpath'))
    {
      $self->error("dbpath: $self->{fields}->{dbpath} does not exist");
      return $self;
    }
  }


  my $lockpath=$self->field('dbpath');
  $lockpath='.' unless ($lockpath);
  $self->{fields}->{lock}=  new Lock ({'lockpath'=>$lockpath,
			       'umask'=>$self->field('umask'),
			       'trials'=>$self->field('timeout'),
			       'keeplocked'=>$self->field('keeplocked'),
			       'lockonly'=>1} );


  $self->error($self->{fields}->{lock}->error());
  unless ($self->{fields}->{lock})
  {
    $self->error("received no valid lock") ;
  }

  local $_ = $self->field('filer');

  #
  # 'require' in stead of 'use',
  # in order to load the specified module only if it is realy required!
  # it is a pitty not to be able to run the program, if one of the
  # modules is not installed, while it is not even needed.
  #
  # Note: Storable was supported, but has proved to be unreliable and
  # has been actively removed for ever and ever.

 fileswitch:
  {
    /^NDBM$/ && do
    {
      require Hashlocker::HLNDBM;
      $self->{'filer'}=new Hashlocker::HLNDBM();
      last fileswitch;
    };

    /^DataDumper$/i && do
    {
      require Hashlocker::HLDataDumper;
      $self->{'filer'}=new Hashlocker::HLDataDumper();
      last fileswitch;
    };

    /^Clone$/i && do
    {
      require Hashlocker::HLClone;
      $self->{'filer'}=new Hashlocker::HLClone();
      last fileswitch;
    };

    $self->error("Invalid filer: $_");
  }

  $self->error($self->{'filer'}->error()) if ($self->{'filer'});
  return $self;
}


sub gendbfilename
{
  my $self=shift;
  my $name=$0;
  $name =~ s/^.*\///;
  $name =~ s/\..*$//;
  $name;
}



sub reading_by
{
  my $self=shift;
  #next parameter is func, pass it on with @_
  my $result='';
  my $dbname=$self->field('dbpath');
  $dbname.='/' if ($dbname);
  $dbname.=$self->field('db');

  unless ($self->field('nolock'))
  {
    $self->{fields}->{lock}->open("$dbname");
    $self->error($self->{fields}->{lock}->error());
  }

  unless ($self->error())
  {
    my $filer =$self->{'filer'};
    $result   =$filer->read($dbname,@_);

    $self->error($filer->error());
    $self->{fields}->{lock}->close() unless ($self->field('nolock'));
  }
  return($result);
}


sub writing_by
{
  my $self=shift;
  my $func=shift;

  my $userhash;
  my $result='';

  my $dbname=$self->field('dbpath');
  $dbname.='/' if ($dbname);
  $dbname.=$self->field('db');

  unless ($self->field('nolock'))
  {
    $self->{fields}->{lock}->open($dbname, {'locktype'=>'LOCK_EX'});
    $self->error($self->{fields}->{lock}->error());
  }

  unless ($self->error())
  {
    my $filer=$self->{'filer'};
    $result=$filer->write($dbname,$func,$self,@_);
    $self->{fields}->{lock}->close() unless ($self->field('nolock'));
    $self->error($filer->error());
  }

  return($result);
}
42;

=head1 Hashlocker.pm

=head1 SYNOPSIS

Provides a locking mechanism for a hash stored on disk. The lock is respected by other Hashlocker programs only.


=head1 DESCRIPTION

Provides a read/write locking mechanism for a hash stored on disk in several possible data formats.

The following 'database types' are supported:

NDBM

This is default for historical reasons. A tied hash is used. Note that the hash can only
can have a one level deep key value structure: the values itself can only contain lexical values.

DataDumper

see Data::Dumper. This goes as deep as your thoughts allow you to. The data is readily readable.
A security consideration: The file is read and passed to an eval(). Anyone who has write access to
this 'data' file, could execute any command on your system under the privliges the programs has.

Clone

Goes as deep as DataDummper, or possibly deeper. As the name suggests, the entire datastructer
can easily be duplicated (cloned). The data are stored encoded (not encrypted) and checksummed.
This might be considered more secure, or at least provides a higher tempering treshold.

=head1 DEPENDENCIES

Data::DataDumper.pm Baseobject.pm Lock.pm Clone.pm


=head1 DETAILS

The constructor of Hashlocker takes a ref to a subroutine as an argument.
Subsequently that subroutine is called back from a read or write context.
The subroutine is (at least) passed  a ref to a locked database as argument to it,
and should return its results.


=head1 PARAMETERS

=over 2

All parameters are optional.
Parameters can be passed as described in L<Baseobject> by passing a ref to a hash.

Par.ex. my $hl = new Hashlocker ({ db=>'foo',
                                  timeout=>2
                                });

=item B<db>

The requested databasename.

Default: The first part of the calling program (foo[.pl]), followed by an extension
corresponding to the 'filer' datatype (foo.DataDumper, foo.Clone, foo.pag/foo.dir).

=item B<dbpath>

The requested path where the database is stored on disk.

Default: $main::localstatedir, which you must have defined yourself.

=item B<filer>

'Database type'. Choose from 'NDBM', 'DataDumper', 'Clone'.

See above documentation for more info.

Recommendation: B<do not use the default>. The default is only choosen for
backward compatibility.

Default: NDBM

=item B<umask>

Use this umask when creating a new database.

=item B<fileperm>

Use these filepermissions for creating a new database.

=item B<nolock>

Don't lock. Rather silly, but might come in handy for testing purposes

Default: 0

=item B<timeout>

A timeout value in seconds can be specified.

Default: 15 seconds. See Lock.pm

=item B<keeplocked>

Set this value to none-zero if you want a lock to continue after a call to
reading_by or writing_by. You might want this if your own program calls one
from the other. Bare in mind that a lock is accepted if it was already active
coming from the same object. In other words: no deadlock occurs where the
object is waiting for itself to finish.
see also: 'lock'

Default: 0

=item B<lock>

Contains the 'lock' object (see Lock.pm) itself, so its fields might be manipulated
directly.

Example:
    $hl->{lock}->{keeplocked}=1;

=back 2

=head1 METHODS

=item B<new()>

Create the object, pass all configuration parameters here.

 Par.ex. my $hl = new Hashlocker ({ db=>'foo',
                                  timeout=>2
                                });

=item B<reading_by()>

Provides the readonly callback, a LOCK_SH, shared or read lock is performed around
the callback. See writing_by below for more details.


=item B<writing_by()>

Write version of reading_by, a LOCK_EX, exclusive or write lock is performed around
the callback.

Both reading_by and writing_by are called with a ref to a subroutine as parameter.
This subroutine is called back, after the lock has been activated.
De callback subroutine receives a ref to the locked dataobject as the first parameter
and optionally all other parameters to writing_by / reading_by are passed as extra parameters.

Note the often practical nested subroutine way of calling these methods:

 my $var;
 $hl->reading_by (
                  sub {
		    my $ref=shift;
                    if (exists ($ref->{$var})
		    {
                       # etc etc
                    }
                  }

    )


=head1 EXAMPLE

 use strict;

 #default location for our dbfile: $main::localstatedir
 our $prefix= '/tmp/test';
 our $localstatedir = "$prefix/var";

 use Hashlocker;

 my $hl = new Hashlocker();
 die $hl->error() if ($hl->error());

 $hl->writing_by(\&schrijf,'extra value');    #something is written to example.db
 die $hl->error() if ($hl->error());

 print $hl->reading_by(\&lees),"\n";          #something is read from example.db
 die $hl->error() if ($hl->error());

 sub schrijf
 {
  my $hashref=shift;
  my $param=shift

  $hashref->{'a'} = 'AA';
  $hashref->{'b'} = 'BB';
  $hashref->{'c'} = 'CC';
  $hashref->{'d'} = $param;
 }

 sub lees
 {
  my $hashref=shift;
  return($hashref->{'b'});
 }
=cut
